/*
    Ming, an SWF output library
    Copyright (C) 2001  Opaque Industries - http://www.opaque.net/

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "shape.h"


struct stateChangeRecord
{
  int flags;
  int moveToX;
  int moveToY;
  int leftFill;
  int rightFill;
  int line;
  /* newstyle not used.. */
};
typedef struct stateChangeRecord *StateChangeRecord;

struct lineToRecord
{
  int dx;
  int dy;
};
typedef struct lineToRecord *LineToRecord;

struct curveToRecord
{
  int controlx;
  int controly;
  int anchorx;
  int anchory;
};
typedef struct curveToRecord *CurveToRecord;

typedef enum
{
  SHAPERECORD_STATECHANGE,
  SHAPERECORD_LINETO,
  SHAPERECORD_CURVETO
} shapeRecordType;

struct shapeRecord
{
  shapeRecordType type;

  union
  {
    StateChangeRecord stateChange;
    LineToRecord lineTo;
    CurveToRecord curveTo;
  } record;
};
typedef struct shapeRecord ShapeRecord;

struct SWFShape_s
{
  struct SWFCharacter_s character;

  ShapeRecord *records;
  int nRecords;
  SWFOutput out;
  int xpos;	/* cursor for using abs. coords in lineTo, curveTo */
  int ypos;
  SWFLineStyle *lines;
  SWFFillStyle *fills;
  byte nLines;
  byte nFills;
  short lineWidth;
  byte isMorph;
  byte isEnded;
};


static void SWFShape_writeShapeRecord(SWFShape shape, ShapeRecord record);


static void writeSWFShapeBlockToMethod(SWFBlock block,
				       SWFByteOutputMethod method, void *data)
{
  SWFOutput out = ((SWFShape)block)->out;
  SWFOutput_writeToMethod(out, method, data);
}


static int completeSWFShapeBlock(SWFBlock block)
{
  SWFShape shape = (SWFShape)block;

  SWFShape_end(shape);

  return SWFOutput_getLength(shape->out);
}


void destroySWFShape(SWFBlock block)
{
  SWFShape shape = (SWFShape)block;
  int i;

  SWFCharacter_clearDependencies(CHARACTER(block));

  for(i=0; i<shape->nFills; ++i)
  {
    SWFMatrix matrix = SWFFillStyle_getMatrix(shape->fills[i]);

    if(matrix != NULL)
      destroySWFMatrix(matrix);

    /* gradients and bitmaps are destroyed separately */

    free(shape->fills[i]);
  }

  if(shape->fills != NULL)
  {
    free(shape->fills);

    for(i=0; i<shape->nLines; ++i)
      free(shape->lines[i]);
  }

  if(shape->lines != NULL)
    free(shape->lines);

  destroySWFRect(CHARACTER(shape)->bounds);
  destroySWFOutput(shape->out);
  free(shape);
}


SWFShape newSWFShape()
{
  SWFShape shape = malloc(sizeof(struct SWFShape_s));

  SWFCharacterInit((SWFCharacter)shape);

  BLOCK(shape)->writeBlock = writeSWFShapeBlockToMethod;
  BLOCK(shape)->complete = completeSWFShapeBlock;
  BLOCK(shape)->dtor = destroySWFShape;
  BLOCK(shape)->type = SWF_DEFINESHAPE3;

  CHARACTERID(shape) = ++SWF_gNumCharacters;

  shape->out = newSWFOutput();
  CHARACTER(shape)->bounds = newSWFRect(0,0,0,0);

  shape->records = NULL;
  shape->lines = NULL;
  shape->fills = NULL;

  shape->nRecords = 0;
  shape->xpos = 0;
  shape->ypos = 0;
  shape->nLines = 0;
  shape->nFills = 0;
  shape->lineWidth = 0;
  shape->isMorph = 0;
  shape->isEnded = 0;

  SWFOutput_writeUInt8(shape->out, 0); /* space for nFillBits, nLineBits */

  return shape;
}


void SWFShape_end(SWFShape shape)
{
  int i;
  byte* buffer;

  if(shape->isEnded)
    return;

  shape->isEnded = TRUE;

  buffer = SWFOutput_getBuffer(shape->out);

  buffer[0] =
    (SWFOutput_numBits(shape->nFills)<<4) + SWFOutput_numBits(shape->nLines);

  for(i=0; i<shape->nRecords; ++i)
  {
    if(i < shape->nRecords-1 ||
       shape->records[i].type != SHAPERECORD_STATECHANGE)
    {
      SWFShape_writeShapeRecord(shape, shape->records[i]);
    }

    free(shape->records[i].record.stateChange); /* all in union are pointers */
  }

  SWFOutput_writeBits(shape->out, 0, 6); /* end tag */
  SWFOutput_byteAlign(shape->out);

  /* addStyleHeader creates a new output and adds the existing one after
     itself- so even though it's called afterwards it's written before,
     as it should be */

  if(BLOCK(shape)->type > 0) /* i.e., shape with style */
    SWFShape_addStyleHeader(shape);

  free(shape->records);
  shape->nRecords = 0;
}


SWFOutput SWFShape_getOutput(SWFShape shape)
{
  return shape->out;
}


SWFFillStyle *SWFShape_getFills(SWFShape shape)
{
  return shape->fills;
}


int SWFShape_getNFills(SWFShape shape)
{
  return shape->nFills;
}


SWFLineStyle *SWFShape_getLines(SWFShape shape)
{
  return shape->lines;
}


int SWFShape_getNLines(SWFShape shape)
{
  return shape->nLines;
}


void SWFShape_setMorphFlag(SWFShape shape)
{
  shape->isMorph = 1;
}


void SWFShape_addStyleHeader(SWFShape shape)
{
  SWFOutput out = newSWFOutput();

  SWFOutput_writeUInt16(out, CHARACTERID(shape));
  SWFOutput_writeRect(out, SWFCharacter_getBounds(CHARACTER(shape)));

  SWFOutput_writeFillStyles(out, shape->fills, shape->nFills, BLOCK(shape)->type);
  SWFOutput_writeLineStyles(out, shape->lines, shape->nLines, BLOCK(shape)->type);

  /* prepend shape->out w/ shape header */
  SWFOutput_setNext(out, shape->out);
  shape->out = out;
}


/*
  ShapeRecords are an intermediate storage so that we don't have to specify
  fill/line types in advance.
*/

#define SHAPERECORD_INCREMENT 32

static ShapeRecord newShapeRecord(SWFShape shape, shapeRecordType type)
{
  if(shape->nRecords % SHAPERECORD_INCREMENT == 0)
  {
    shape->records = realloc(shape->records,
			     sizeof(ShapeRecord) *
			     (shape->nRecords + SHAPERECORD_INCREMENT));
  }

  switch(type)
  {
    case SHAPERECORD_STATECHANGE:
    {
      StateChangeRecord change = calloc(1,sizeof(struct stateChangeRecord));
      shape->records[shape->nRecords].record.stateChange = change;
      break;
    }
    case SHAPERECORD_LINETO:
    {
      LineToRecord lineTo = calloc(1,sizeof(struct lineToRecord));
      shape->records[shape->nRecords].record.lineTo = lineTo;
      break;
    }
    case SHAPERECORD_CURVETO:
    {
      CurveToRecord curveTo = calloc(1,sizeof(struct curveToRecord));
      shape->records[shape->nRecords].record.curveTo = curveTo;
      break;
    }
  }

  shape->records[shape->nRecords].type = type;
  return shape->records[shape->nRecords++];
}


void SWFShape_writeShapeRecord(SWFShape shape, ShapeRecord record)
{
  SWFOutput out = shape->out;

  switch(record.type)
  {
    case SHAPERECORD_STATECHANGE:
    {
      int flags = record.record.stateChange->flags;

      if(flags == 0)
	return;

      SWFOutput_writeBits(out, flags, 6);

      if(flags & SWF_SHAPE_MOVETOFLAG)
      {
	int x = record.record.stateChange->moveToX;
	int y = record.record.stateChange->moveToY;
        int nBits = max(SWFOutput_numSBits(x), SWFOutput_numSBits(y));

        SWF_assert(nBits<32);
        SWFOutput_writeBits(out, nBits, 5);
        SWFOutput_writeSBits(out, x, nBits);
        SWFOutput_writeSBits(out, y, nBits);
      }

      if(flags & SWF_SHAPE_FILLSTYLE0FLAG)
      {
        SWFOutput_writeBits(out, record.record.stateChange->leftFill,
			    SWFOutput_numBits(shape->nFills));
      }

      if(flags & SWF_SHAPE_FILLSTYLE1FLAG)
      {
        SWFOutput_writeBits(out, record.record.stateChange->rightFill,
			    SWFOutput_numBits(shape->nFills));
      }

      if(flags & SWF_SHAPE_LINESTYLEFLAG)
      {
        SWFOutput_writeBits(out, record.record.stateChange->line,
			    SWFOutput_numBits(shape->nLines));
      }

      /* newstyle's never used.  But this is what it looks like:

      if(flags & SWF_SHAPE_NEWSTYLEFLAG)
      {
        SWFOutput_writeFillStyles(shape->out, shape->fills, shape->nFills,
				  BLOCK(shape)->type);

        SWFOutput_writeLineStyles(shape->out, shape->lines, shape->nLines,
				  BLOCK(shape)->type);

	SWFOutput_writeBits(shape->out, SWFOutput_numBits(shape->nFills), 4);
        SWFOutput_writeBits(shape->out, SWFOutput_numBits(shape->nLines), 4);
      }

      */

      break;
    }

    case SHAPERECORD_LINETO:
    {
      int nBits;
      int dx = record.record.lineTo->dx;
      int dy = record.record.lineTo->dy;

      SWFOutput_writeBits(out, 3, 2); /* straight edge */

      if(dx==0)
      {
	nBits = SWFOutput_numSBits(dy);
	SWF_assert(nBits<18);
	SWFOutput_writeBits(out, nBits-2, 4);
	SWFOutput_writeBits(out, 1, 2); /* vertical line */
	SWFOutput_writeSBits(out, dy, nBits);
      }
      else if(dy==0)
      {
	nBits = SWFOutput_numSBits(dx);
	SWF_assert(nBits<18);
	SWFOutput_writeBits(out, nBits-2, 4);
	SWFOutput_writeBits(out, 0, 2); /* horizontal line */
	SWFOutput_writeSBits(out, dx, nBits);
      }
      else
      {
	nBits = max(SWFOutput_numSBits(dx), SWFOutput_numSBits(dy));
	SWF_assert(nBits<18);
	SWFOutput_writeBits(out, nBits-2, 4);
	SWFOutput_writeBits(out, 1, 1); /* general line */
	SWFOutput_writeSBits(out, dx, nBits);
	SWFOutput_writeSBits(out, dy, nBits);
      }

      break;
    }

    case SHAPERECORD_CURVETO:
    {
      int controlx = record.record.curveTo->controlx;
      int controly = record.record.curveTo->controly;
      int anchorx = record.record.curveTo->anchorx;
      int anchory = record.record.curveTo->anchory;

      int nBits = max(max(SWFOutput_numSBits(controlx),
			  SWFOutput_numSBits(controly)),
		      max(SWFOutput_numSBits(anchorx),
			  SWFOutput_numSBits(anchory)));

      if(nBits<2)
	nBits = 2;

      SWF_assert(nBits<18);

      SWFOutput_writeBits(out, 2, 2); /* curved edge */
      SWFOutput_writeBits(out, nBits-2, 4);
      SWFOutput_writeSBits(out, controlx, nBits);
      SWFOutput_writeSBits(out, controly, nBits);
      SWFOutput_writeSBits(out, anchorx, nBits);
      SWFOutput_writeSBits(out, anchory, nBits);

      break;
    }

    default:
      SWF_error("Unknown shapeRecordType");
  }
}


/* x,y relative to shape origin */
void SWFShape_drawScaledLineTo(SWFShape shape, int x, int y)
{
  SWFShape_drawScaledLine(shape, x-shape->xpos, y-shape->ypos);
}


void SWFShape_drawScaledLine(SWFShape shape, int dx, int dy)
{
  ShapeRecord record;

  if(shape->isEnded)
    return;

  if(dx == 0 && dy == 0)
    return;

  record = newShapeRecord(shape, SHAPERECORD_LINETO);

  SWF_assert(SWFOutput_numSBits(dx) < 18);
  SWF_assert(SWFOutput_numSBits(dy) < 18);

  record.record.lineTo->dx = dx;
  record.record.lineTo->dy = dy;

  shape->xpos += dx;
  shape->ypos += dy;

  SWFRect_includePoint(SWFCharacter_getBounds(CHARACTER(shape)),
		       shape->xpos, shape->ypos, shape->lineWidth);
}


void SWFShape_drawScaledCurveTo(SWFShape shape, int controlx, int controly,
				int anchorx, int anchory)
{
  SWFShape_drawScaledCurve(shape, controlx-shape->xpos, controly-shape->ypos,
			   anchorx-controlx, anchory-controly);
}


void SWFShape_drawScaledCurve(SWFShape shape, int controldx, int controldy,
			int anchordx, int anchordy)
{
  ShapeRecord record;

  if(shape->isEnded)
    return;

  if(controldx == 0 && controldy == 0 && anchordx == 0 && anchordy == 0)
    return;

  record = newShapeRecord(shape, SHAPERECORD_CURVETO);

  record.record.curveTo->controlx = controldx;
  record.record.curveTo->controly = controldy;
  record.record.curveTo->anchorx = anchordx;
  record.record.curveTo->anchory = anchordy;

  if(SWFOutput_numSBits(controldx) >= 18 ||
     SWFOutput_numSBits(controldy) >= 18 ||
     SWFOutput_numSBits(anchordx) >= 18 ||
     SWFOutput_numSBits(anchordy) >= 18)
    SWF_error("Curve parameters too large");

  /* including the control point is sloppy, but safe.. */

  shape->xpos += controldx;
  shape->ypos += controldy;

  SWFRect_includePoint(SWFCharacter_getBounds(CHARACTER(shape)),
		       shape->xpos, shape->ypos, shape->lineWidth);

  shape->xpos += anchordx;
  shape->ypos += anchordy;

  SWFRect_includePoint(SWFCharacter_getBounds(CHARACTER(shape)),
		       shape->xpos, shape->ypos, shape->lineWidth);
}


#define STYLE_INCREMENT 4

static int SWFShape_addLineStyle(SWFShape shape, unsigned short width,
				 byte r, byte g, byte b, byte a)
{
  if(shape->nLines % STYLE_INCREMENT == 0)
  {
    shape->lines = realloc(shape->lines,
			   (shape->nLines+STYLE_INCREMENT) *
			   sizeof(SWFLineStyle));
  }

  shape->lines[shape->nLines] = newSWFLineStyle(width, r, g, b, a);

  return ++shape->nLines;
}


/* if the current shape record isn't a style change record, add one */
static ShapeRecord addStyleRecord(SWFShape shape)
{
  if(shape->nRecords > 0 &&
     shape->records[shape->nRecords-1].type == SHAPERECORD_STATECHANGE)
  {
    return shape->records[shape->nRecords-1];
  }
  else
    return newShapeRecord(shape, SHAPERECORD_STATECHANGE);
}


void SWFShape_setLineStyle(SWFShape shape, unsigned short width,
			   byte r, byte g, byte b, byte a)
{
  ShapeRecord record;
  int line;

  if(shape->isEnded)
    return;

  for(line=0; line<shape->nLines; ++line)
  {
    if(SWFLineStyle_equals(shape->lines[line], width, r, g, b, a))
      break;
  }

  if(line == shape->nLines)
    line = SWFShape_addLineStyle(shape, width, r, g, b, a);
  else
    ++line;

  if(width == 0)
    shape->lineWidth = 0;
  else
    shape->lineWidth = (SWFLineStyle_getWidth(shape->lines[line-1])+1)/2;

  if(shape->isMorph)
    return;

  record = addStyleRecord(shape);

  record.record.stateChange->line = line;
  record.record.stateChange->flags |= SWF_SHAPE_LINESTYLEFLAG;
}


/* fill 0 is no fill, so set fill->idx to one more than the shape's fill index */
static SWFFillStyle addFillStyle(SWFShape shape, SWFFillStyle fill)
{
  int i;

  for(i=0; i<shape->nFills; ++i)
  {
    if(SWFFillStyle_equals(fill, shape->fills[i]))
    {
      free(fill);
      return shape->fills[i];
    }
  }

  if(shape->isEnded)
  {
    SWFFill_setIdx(fill, 0);
    return fill;
  }

  if(shape->nFills%STYLE_INCREMENT == 0)
  {
    shape->fills = realloc(shape->fills,
			   (shape->nFills+STYLE_INCREMENT) *
			   sizeof(SWFFillStyle));
  }

  SWFFill_setIdx(fill, shape->nFills+1);

  shape->fills[shape->nFills] = fill;
  ++shape->nFills;

  return fill;
}


SWFFillStyle SWFShape_addSolidFillStyle(SWFShape shape,
					byte r, byte g, byte b, byte a)
{
  return addFillStyle(shape, newSWFSolidFillStyle(r, g, b, a));
}


SWFFillStyle SWFShape_addGradientFillStyle(SWFShape shape,
					   SWFGradient gradient, byte flags)
{
  return addFillStyle(shape, newSWFGradientFillStyle(gradient, flags));
}


SWFFillStyle SWFShape_addBitmapFillStyle(SWFShape shape,
					 SWFBitmap bitmap, byte flags)
{
  SWFCharacter_addDependency((SWFCharacter)shape, (SWFBlock)bitmap);

  return addFillStyle(shape, newSWFBitmapFillStyle(bitmap, flags));
}


void SWFShape_setLeftFillStyle(SWFShape shape, SWFFillStyle fill)
{
  ShapeRecord record;

  if(shape->isEnded || shape->isMorph)
    return;

  record = addStyleRecord(shape);

  if(fill != NOFILL)
  {
    if(SWFFill_getIdx(fill) > shape->nFills)
      SWF_error("Invalid fill idx");

    record.record.stateChange->leftFill = SWFFill_getIdx(fill);
  }
  else
    record.record.stateChange->leftFill = 0;

  record.record.stateChange->flags |= SWF_SHAPE_FILLSTYLE0FLAG;
}


void SWFShape_setRightFillStyle(SWFShape shape, SWFFillStyle fill)
{
  ShapeRecord record;

  if(shape->isEnded || shape->isMorph)
    return;

  record = addStyleRecord(shape);

  if(fill != NOFILL)
  {
    if(SWFFill_getIdx(fill) > shape->nFills)
      SWF_error("Invalid fill idx");

    record.record.stateChange->rightFill = SWFFill_getIdx(fill);
  }
  else
    record.record.stateChange->rightFill = 0;

  record.record.stateChange->flags |= SWF_SHAPE_FILLSTYLE1FLAG;
}


/* move pen relative to shape origin */
void SWFShape_moveScaledPenTo(SWFShape shape, int x, int y)
{
  ShapeRecord record;

  if(shape->isEnded)
    return;

  record = addStyleRecord(shape);

  record.record.stateChange->moveToX = shape->xpos = x;
  record.record.stateChange->moveToY = shape->ypos = y;

  record.record.stateChange->flags |= SWF_SHAPE_MOVETOFLAG;

  if(shape->nRecords == 0 ||
     (shape->nRecords == 1 &&
      shape->records[0].type == SHAPERECORD_STATECHANGE))
  {
    SWFRect_setBounds(SWFCharacter_getBounds(CHARACTER(shape)), x, x, y, y);
  }
}


void SWFShape_moveScaledPen(SWFShape shape, int x, int y)
{
  SWFShape_moveScaledPenTo(shape, shape->xpos+x, shape->ypos+y);
}


int SWFShape_getScaledPenX(SWFShape shape)
{
  return shape->xpos;
}


int SWFShape_getScaledPenY(SWFShape shape)
{
  return shape->ypos;
}


/* yes, this is a total hack. */

#include "read.c"

void SWFShape_drawScaledGlyph(SWFShape shape,
			      SWFFont font, unsigned char c, int size)
{
  byte *p = SWFFont_findCharacterGlyph(font, c);
  byte **f = &p;

  int moveBits, x, y;
  int straight, numBits;

  /* moveTos in the record are absolute, but we want to draw from the current
     location. grr. */

  int startX = shape->xpos;
  int startY = shape->ypos;

  byteAlign();

  if(readBits(f, 4) != 1) /* fill bits */
    SWF_error("SWFShape_drawGlyph: bad file format (was expecting fill bits = 1)");

  if(readBits(f, 4) != 0) /* line bits */
    SWF_error("SWFShape_drawGlyph: bad file format (was expecting line bits = 0)");

  /* now we get to parse the shape commands.  Oh boy.
     the first one will be a non-edge block- grab the moveto loc */

  readBits(f, 6); /* type 0, etc. */

  moveBits = readBits(f, 5);
  x = startX + readSBits(f, moveBits);
  y = startY + readSBits(f, moveBits);

  SWFShape_moveScaledPenTo(shape, x*size/1024, y*size/1024);

  if(readBits(f, 1) != 1) /* fill1 = 1 */
    SWF_error("SWFShape_drawGlyph: bad file format (was expecting fill1 = 1)");

  /* translate the glyph's shape records into drawing commands */

  for(;;)
  {
    if(readBits(f, 1) == 0)
    {
      /* it's a moveTo or a shape end */

      if(readBits(f, 5) == 0)
	break;

      moveBits = readBits(f, 5);
      x = startX + readSBits(f, moveBits);
      y = startY + readSBits(f, moveBits);

      SWFShape_moveScaledPenTo(shape, x*size/1024, y*size/1024);

      continue;
    }

    straight = readBits(f, 1);
    numBits = readBits(f, 4)+2;

    if(straight==1)
    {
      if(readBits(f, 1)) /* general line */
      {
	x += readSBits(f, numBits);
	y += readSBits(f, numBits);
      }
      else
      {
	if(readBits(f, 1)) /* vert = 1 */
	  y += readSBits(f, numBits);
	else
	  x += readSBits(f, numBits);
      }

      SWFShape_drawScaledLineTo(shape, x*size/1024, y*size/1024);
    }
    else
    {
      int controlX = readSBits(f, numBits);
      int controlY = readSBits(f, numBits);
      int anchorX = readSBits(f, numBits);
      int anchorY = readSBits(f, numBits);

      SWFShape_drawScaledCurveTo(shape,
				 (x+controlX)*size/1024,
				 (y+controlY)*size/1024,
				 (x+controlX+anchorX)*size/1024,
				 (y+controlY+anchorY)*size/1024);

      x += controlX + anchorX;
      y += controlY + anchorY;
    }
  }

  /* no idea where the pen was left */
  SWFShape_moveScaledPenTo(shape, startX, startY);
}
